package com.blazinggames.urcl.text;

// Java imports
import java.awt.*;
import java.awt.event.KeyEvent;
import java.util.Random;

// URCL imports
import com.blazinggames.urcl.data.*;
import com.blazinggames.urcl.ui.*;
import com.blazinggames.urcl.ui.playfield.*;

/**
 * Ultimate Retro Common Library
 * Copyright (C) 2001, 2003-2004 Blazing Games Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ------------------------------------------------
 * 
 * A display class specifically designed for simulating old teletype games.
 * 
 * @author Billy D. Spelchan
 * @version OSR1.00
 */
public class TextDisplay
extends Display
implements Runnable
{
// *********************
// ***** CONSTANTS *****
// *********************
    private static final Point DEF_FONTSIZE = new Point(16,16);
    
    protected static final String TELETYPE_IMAGE = "TTFONT";
    protected static final String LCD_IMAGE = "TTLCDFONT";
    protected static final String FONTSIZE = "TTFONTSIZE";
    protected static final String CHARDELAY = "TTCHARDELAY";
    protected static final String LINEDELAY = "TTLINEDELAY";

// *********************
// ***** VARIABLES *****
// *********************

    // Page display variables
    Image _ttimage;
    Image _lcdimage;
    TextTile _texttile;
    TextTile _lcdtile;
    Playfield _page;
    int _fontWidth = 16;
    int _fontHeight = 16;
    int _pageWidth;
    int _pageHeight;
    int _putPosition;
    Rectangle _fullViewport;
    Rectangle _bottomViewport;
    boolean _fullRender;
    boolean _scrolling = false;
    int _scrollPosition;
    
    // Printing delay variables (will be in config)
    int _delayBetweenChars = 50;
    int _delayBetweenLines = 250;
    
    // input system
    Playfield _inputLine;
    Rectangle _inputViewport;
    int _inputPosition;
    boolean _waitingForInput = false;

    // Random number aids
    Random _random;
    
    // debug
    Font _frFont = new Font ("Serif", 0, 12);
    long _lastFrame;
    
    int _tick_count;
    Thread _gamethread = null;

    protected boolean _canQuit = false;
    
// ***********************
// ***** CONSTRUCTOR *****
// ***********************
    
    public TextDisplay()
    {
        _random = new Random();
    }


// ***********************************
// ***** METHODS - Basic running *****
// ***********************************

	/**
	 * sets up the playfield that the text display uses.
	 */
    protected void preparePage()
    {
        _page = new Playfield();
        Container c = _displayHandler.getContainer();
        Dimension d = c.getSize();
        _pageWidth = d.width / _fontWidth;
        _pageHeight = d.height / _fontHeight;
        _page.setTileset(2, _fontWidth, _fontHeight);
        _page.setTileRenderer(0, _texttile);
        _page.setPlayfield(_pageWidth, _pageHeight);
        _page.clear(0, 32);
        _fullViewport = new Rectangle (0,0, _pageWidth * _fontWidth, _pageHeight * _fontHeight - _fontHeight);
        _bottomViewport = new Rectangle(0, _pageHeight*_fontHeight-(_fontHeight*2), _pageWidth*_fontWidth,_fontHeight);
        _fullRender = true;

        _inputLine = new Playfield();
        _inputLine.setTileset(2, _fontWidth, _fontHeight);
        _inputLine.setTileRenderer(0, _lcdtile);
        _inputLine.setPlayfield(_pageWidth, 1);
        _inputLine.clear(0, 32);
    	_inputViewport = new Rectangle(0, 0, _pageWidth*_fontWidth,_fontHeight);
    }

	/**
	 * initializes the text display by loading all the graphics and calling 
	 * preparePage.
	 * 
	 * @param dh the Display Handler that handles this display.
	 */    
    public boolean initialize(DisplayHandler dh)
    {
        super.initialize(dh);
        dh.displayEvent(this, "LOADING 34");
        Point ptemp = _config.getPointProperty(FONTSIZE, DEF_FONTSIZE);
        _fontWidth = ptemp.x;
        _fontHeight = ptemp.y;
        String temp = _config.getStringProperty(TELETYPE_IMAGE, "res/urv1/Teletype2.GIF");
        System.out.println("teletype = " + temp);
        _res.addImage(TELETYPE_IMAGE, temp);
        _ttimage = _res.getImage(TELETYPE_IMAGE);
        _texttile = new TextTile(_ttimage, _fontWidth, _fontHeight);
        dh.displayEvent(this, "LOADING 67");
        temp = _config.getStringProperty(LCD_IMAGE, "res/urv1/lcd.gif");
        System.out.println("lcd = " + temp);
        _res.addImage(LCD_IMAGE, temp);
        _lcdimage = _res.getImage(LCD_IMAGE);
        _lcdtile = new TextTile(_lcdimage, _fontWidth, _fontHeight);
        _delayBetweenChars = _config.getIntProperty(CHARDELAY, 50);
        _delayBetweenLines = _config.getIntProperty(LINEDELAY, 250);
        preparePage();
        dh.displayEvent(this, "LOADING 100");
        return true;
    }
    
    /**
     * starts the game thread.
     */
    public boolean begin()
    {
        _fullRender = true;
        if (_gamethread==null)
        {
            _gamethread = new Thread(this);
            _gamethread.start();
        }
        return super.begin();
    }
    
    /**
     * ends the game thread
     */
    public boolean end()
    {
_displayHandler.log("end() starting");
        if (_gamethread != null)
        {
//            _gamethread.stop();
            _gamethread = null;
        }
_displayHandler.log("end() ending");
        return super.end();
    }
    
    /**
     * Designed in case at a future date I want to combine a bunch of games
     * together so that the text games know that they should quit when done.
     * 
     * @param b true if want to enable quit flag
     */
    public void enableQuit(boolean b)
    {
        _canQuit = b;
    }

// ************************************
// ***** METHODS - Event handling *****
// ************************************

	/**
	 * Handles keyboard input by updating LCD portion of display if waiting
	 * for input otherwise ignores key press.
	 * 
	 * @param code keycode of key pressed
	 * @param ch character of key pressed
	 */
    public void key_up(int code, char ch)
    {
        if (_waitingForInput)
        {
            if (code == KeyEvent.VK_ENTER)
            {
                _waitingForInput = false;
            }
            else if ((code == KeyEvent.VK_BACK_SPACE) || (code == KeyEvent.VK_DELETE))
            {
                _inputLine.setTile(_inputPosition, 0, 0, 32);
                _inputPosition = Math.max(0, _inputPosition - 1);
                _inputLine.setTile(_inputPosition, 0, 0, 127);
                _displayHandler.refreshDisplay(this);                
            }
            else if ((ch > 31) && (ch < 128) && (_inputPosition < _pageWidth))
            {
                _inputLine.setTile(_inputPosition, 0, 0, ch);
                ++_inputPosition;
                _inputLine.setTile(_inputPosition, 0, 0, 127);
                _displayHandler.refreshDisplay(this);
            }
        }
    }
    
    /**
     * handles tick by animating display
     */
    public void tick(int ms)
    {
        ++_tick_count;
        if (_scrolling)
        {
            ++_fullViewport.y;
            if (_fullViewport.y>=_fontHeight)
            {
                _page.shift(0, -1, (short)0, (short)32);
                _fullViewport.y = 0;
                _scrolling = false;
            }
            _fullRender = true;
        }
//        put( (char)_tick_count );
//        _displayHandler.log("Tick - " + _tick_count);
        _displayHandler.refreshDisplay(this);
    }


// ************************************************
// ***** METHODS - Display (and display aids) *****
// ************************************************

	/**
	 * draws the display.
	 * 
	 * @param g graphics context
	 * @param comp component drawing occurring in
	 */    
    public synchronized void render(Graphics g, Component comp)
    {
        if (_fullRender)
        {
            _fullRender = false;
            _page.render(g, comp, 0, 0, _fullViewport);            
        }
        else
        {
            _page.render(g, comp, _bottomViewport.x, _bottomViewport.y, _bottomViewport);
        }
        _inputLine.render(g, comp, 0, _pageHeight*_fontHeight-_fontHeight, _inputViewport);
    }



// ***************************************
// ***** METHODS - Printing commands *****
// ***************************************

	/**
	 * Draws a character on the display.
	 * 
	 * @param ch character to draw to display
	 */
    public void put(char ch)
    {
        if ((ch & 127) == 13)
            _putPosition = _pageWidth;
        else
        {
            _page.setTile(_putPosition, _pageHeight-2, 0, ch & 127);
            _displayHandler.refreshDisplay(this);
            try {Thread.sleep(_delayBetweenChars);}catch(Exception ignore){;}
        }
        ++_putPosition;
        if (_putPosition >= _pageWidth)
        {
//            _displayHandler.log("Char at 0,0 is " + _page.getTile(0,0) + "/" + _page.getAttribute(0,0));
            _scrolling = true;
            do
            {
                try {Thread.sleep(100);}catch(Exception ignore){;}
            }
            while (_scrolling == true);
            _putPosition = 0;
            try {Thread.sleep(_delayBetweenLines);}catch(Exception ignore){;}
        }
    }

	/**
	 * Sends the passed line of text to the display.
	 * 
	 * @param s Line of text to be displayed
	 */
    public void print(String s)
    {
        int l = s.length();
        for (int cntr = 0; cntr < l; ++cntr)
        {
            put(s.charAt(cntr));
        }
    }
    
	/**
	 * Goes to the next line of the display.
	 */
    public void println()
    {
        put((char)13);
    }
    
	/**
	 * Sends the passed line of text to the display then goes to the next
	 * line of the display.
	 * 
	 * @param s Line of text to be displayed
	 */
    public void println(String s)
    {
        print(s);
        put((char)13);
    }

	/**
	 * gets input from the lcd.
	 * 
	 * @return whatever the user typed into the lcd display
	 */
    public String input()
    {
        _inputLine.setTile(_inputPosition, 0, 0, 127);
        _waitingForInput = true;
        do
        {
            try { Thread.sleep(100); } catch (Exception ignore) {;}
        }
        while (_waitingForInput);
        
        StringBuffer sb = new StringBuffer();
        for (int cntr = 0; cntr < _inputPosition; ++cntr)
        {
            sb.append((char)_inputLine.getAttribute(cntr, 0));
        }
        _inputLine.clear(0,32);
        _inputPosition = 0;
        return sb.toString();
    }


// ***************************************************
// ***** METHODS - Random number generation aids *****
// ***************************************************

	/**
	 * Generates a random number between 1 and n inclusive.
	 * 
	 * @param n Upper range of random number to return
	 * @return random number
	 */
    public int rand(int n)
    {
        return (((_random.nextInt() & Integer.MAX_VALUE) % n) + 1);
    }
    
    /**
     * Generates a random number between low and high inclusive.
     * 
     * @param low low number of range
     * @param high high number of range
     * @return random number
     */
    public int rand(int low, int high)
    {
        int r = high - low + 1;
        if (r <= 1)
            return low;
        return (((_random.nextInt() & Integer.MAX_VALUE) % r) + low);
    }
    
    /**
     * Simulates the rolling of a d sided die rolled n times and then mod
     * added to the result (nDd+mod for RPGers).
     * 
     * @param n number of dice to roll
     * @param d sides each die has
     * @param mod modifier number
     * @return dice roll
     */
    public int roll(int n, int d, int mod)
    {
        int ret = 0;
        for (int cntr = 0; cntr < n; ++cntr)
        {
            ret += rand(d);
        }
        return (ret+mod);
    }

// *************************************************
// ***** METHODS - From Runnable (for testing) *****
// *************************************************

	/**
	 * THIS IS JUST FOR TESTING!
	 */
    public void run()
    {
        int cntr;
        
        println("First, lets start with the put test.");
        for (cntr = 0; cntr < 128; ++cntr)
            put((char)cntr);
        println();
        println("Now, some string tests:");
        println("THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG");
        println("the quick brown fox jumps over the lazy dog");
        println("1234567890!@#$%^&*()[]{}\\|/=+-_;:'\"<,?/>.");

        println();
        print("Type Something: ");
        String s = input();
        println(s);
        print("Again: ");
        s=input();
        println(s);
        println("One last time: ");
        s=input();
        println(s);
        println();
        
        println("DONE!!!");
    }
}
